unit rkSmartGfx;

// rkSmartGfx by Roy Magne Klever
//  2011 by Roy Magne Klever. All rights reserved
//
// This file is not distributable without permission by Roy Magne Klever
// WEB: www.rmklever.com
// Mail: roymagne@rmklever.com
//
// September 2011
//
// You are free to use this in any compiled source projects

interface

uses
  rkColorUtil, Windows, Graphics, Math, ExtDlgs, Jpeg, ExtCtrls, Types;

  procedure ShadeGradientHorz(Bmp: TBitmap; Back, Color: TColor; Reverse, Transparent: Boolean);
  procedure ShadeGradientVert(Bmp: TBitmap; Back, Color: TColor; Reverse, Transparent: Boolean);
  function CalcImgSize(W, H, TW, TH: Integer): TPoint;
  procedure fxBoxBlur(Src: TBitmap; radH, radV, rep: Integer);
  procedure fxBCS(Dest: TBitmap; B, C, S: Integer);
  procedure fxHistCalc(Src: TBitmap; var HistR, HistG, HistB: THistogram);
  procedure fxHistEqu(Src: TBitmap; Z: Single);
  procedure fxExposure(Src: TBitmap; K: Single);
  procedure fxGrayScale(Src: TBitmap; Kind: Byte);
  procedure fxSpray(Src: TBitmap; Amount: Integer);
  procedure fxSplitLight(Src: TBitmap; Amount: Integer);
  procedure fxSplitBlur(Src: TBitmap; Amount: Integer);
  procedure fxSplitColor(Src: TBitmap; Amount: Integer);
  procedure fxMosaic(Src: TBitmap; Amount: Integer);
  procedure fxHorzRoll(Src: TBitmap; Amount: Integer);
  procedure fxRoll(Src: TBitmap; RollH, RollV: Integer);
  procedure fxAutoYHist(Src: TBitmap; Z: Single);
  procedure fxAutoContrast(Src : TBitmap);
  procedure fxSoftFilter(Src: TBitmap; Amount: Integer);

implementation

procedure ShadeGradientHorz(Bmp: TBitmap; Back, Color: TColor; Reverse, Transparent: Boolean);
var
  X, Y, w, h, j, w1: Integer;
  Row: PRGBArray;
  slMain, slSize: Integer;
  R, G, B, a: Byte;
begin
  Bmp.Canvas.Pen.Style:= psClear;
  Bmp.Canvas.Brush.Color:= Back;
  if not Transparent then
    Bmp.Canvas.Rectangle(Rect(0, 0, Bmp.Width + 1, Bmp.Height + 1));
  w := Bmp.Width;
  h := Bmp.Height;
  Color := ColorToRGB(Color);
  R := Byte(Color);
  G := Byte(Color shr 8);
  B := Byte(Color shr 16);
  w1 := w - 1;
  w := (w shr 1) + (w and 1);
  slMain := Integer(Bmp.ScanLine[0]);
  slSize := Integer(Bmp.ScanLine[1]) - slMain;
  for Y := 0 to h - 1 do
  begin
    if Reverse then
      j := 255 - MulDiv(255, Y, h)
    else
      j := MulDiv(255, Y + 1, h);
    Row := PRGBArray(slMain);
    for X := 0 to w - 1 do
    begin
      a := 255 - MulDiv(j, X, w);
      Row[X].R := a * (Row[X].R - R) shr 8 + R;
      Row[X].G := a * (Row[X].G - G) shr 8 + G;
      Row[X].B := a * (Row[X].B - B) shr 8 + B;
      if (X < (w1 - X)) then
      begin
        Row[w1 - X].R := a * (Row[w1 - X].R - R) shr 8 + R;
        Row[w1 - X].G := a * (Row[w1 - X].G - G) shr 8 + G;
        Row[w1 - X].B := a * (Row[w1 - X].B - B) shr 8 + B;
      end;
    end;
    slMain := slMain + slSize;
  end;
end;

procedure ShadeGradientVert(Bmp: TBitmap; Back, Color: TColor; Reverse, Transparent: Boolean);
var
  X, Y, w, h, j, w1, h1, x1: Integer;
  Row, Row2: PRGBArray;
  slMain, slOther, slSize: Integer;
  R, G, B, a: Byte;
begin
  Bmp.Canvas.Pen.Style:= psClear;
  Bmp.Canvas.Brush.Color:= Back;
  if not Transparent then
    Bmp.Canvas.Rectangle(Rect(0, 0, Bmp.Width + 1, Bmp.Height + 1));
  w := Bmp.Width;
  h := Bmp.Height;
  Color := ColorToRGB(Color);
  R := Byte(Color);
  G := Byte(Color shr 8);
  B := Byte(Color shr 16);
  slMain := Integer(Bmp.ScanLine[0]);
  slOther := Integer(Bmp.ScanLine[h - 1]);
  slSize := Integer(Bmp.ScanLine[1]) - slMain;
  if Reverse then
    x1:= 1
  else
    x1:= -1;
  h1 := h - 1;
  h := (h shr 1) + (h and 1);
  for Y := 0 to h - 1 do
  begin
    j := MulDiv(255, Y + 1, h);
    Row := PRGBArray(slMain);
    Row2 := PRGBArray(slOther);
    if Reverse then
      w1:= w - 1
    else
      w1:= 0;
    for X := 0 to w - 1 do
    begin
      a := 255 - MulDiv(j, w1, w);
      w1:= w1 - x1;
      Row[X].R := a * (Row[X].R - R) shr 8 + R;
      Row[X].G := a * (Row[X].G - G) shr 8 + G;
      Row[X].B := a * (Row[X].B - B) shr 8 + B;
      if (Y < (h1 - Y)) then
      begin
        Row2[X].R := a * (Row2[X].R - R) shr 8 + R;
        Row2[X].G := a * (Row2[X].G - G) shr 8 + G;
        Row2[X].B := a * (Row2[X].B - B) shr 8 + B;
      end;
    end;
    slMain := slMain + slSize;
    slOther := slOther - slSize;
  end;
end;

function CalcImgSize(W, H, TW, TH: Integer): TPoint;
begin
  Result.X := 0;
  Result.Y := 0;
  if (w < tw) and (h < th) then
  begin
    Result.X := w;
    Result.Y := h;
  end
  else if (w = 0) or (h = 0) then
    Exit
  else
  begin
    if w > h then
    begin
      if w < tw then
        tw := w;
      Result.X := tw;
      Result.Y := Trunc(tw * h / w);
      if Result.Y > th then
      begin
        Result.Y := th;
        Result.X := Trunc(th * w / h);
      end;
    end
    else
    begin
      if h < th then
        th := h;
      Result.Y := th;
      Result.X := Trunc(th * w / h);
      if Result.X > tw then
      begin
        Result.X := tw;
        Result.Y := Trunc(tw * h / w);
      end;
    end;
  end;
end;

procedure fxBoxBlur(Src: TBitmap; radH, radV, rep: Integer);
var
  i, j, w, h, x, y, k, t: Integer;
  F: Integer;
  SL: array of PRGBArray;
  SL2: PRGBArray;
  LUT: array of TiRGB;
begin
  if (radH = 0) and (radV = 0) then
    Exit;
  w := Src.Width - 1;
  h := Src.Height - 1;
  SetLength(LUT, Max(w, h) + 1 + (Max(radH, radV) * 2));
  SetLength(SL, Src.Height);
  SL[0] := (Src.Scanline[0]);
  i := Integer(Src.Scanline[1]) - Integer(SL[0]);
  for j := 1 to Src.Height - 1 do
    SL[j] := PRGBArray(Integer(SL[j - 1]) + i);
  SL2:= SL[1];
  if radH > 0 then
  begin
    t:= radH + radH - 1;
    F := Round((1 / t) * $10000);
    for y := 0 to h do      // Horizontal blur
    begin
      for j := 1 to rep do
      begin
        k := 0;
        for i := -radH to w + radH do
        begin
          if (i > 0) and (i <= w) then
            k := i;
          LUT[i + radH].R := SL2[k].R + LUT[i + radH - 1].R;
          LUT[i + radH].G := SL2[k].G + LUT[i + radH - 1].G;
          LUT[i + radH].B := SL2[k].B + LUT[i + radH - 1].B;
        end;
        for x := 0 to w do
        begin
          SL2[x].R := ((LUT[x + t].R - LUT[x].R) * F) shr 16;
          SL2[x].G := ((LUT[x + t].G - LUT[x].G) * F) shr 16;
          SL2[x].B := ((LUT[x + t].B - LUT[x].B) * F) shr 16;
        end;
        SL2 := SL[y];
      end;
    end;
  end;
  if radV > 0 then
  begin
    t:= radV + radV - 1;
    F := Round((1 / t) * $10000);
    for x := 0 to w do      // Vertical blur
    begin
      for j := 1 to rep do
      begin
        k := 0;
        for i := -radV to h + radV do
        begin
          if (i > 0) and (i < h) then
            k := i;
          LUT[i + radV].R := SL[k][x].R + LUT[i + radV - 1].R;
          LUT[i + radV].G := SL[k][x].g + LUT[i + radV - 1].G;
          LUT[i + radV].B := SL[k][x].B + LUT[i + radV - 1].B;
        end;
        for y := 0 to h do
        begin
          SL[y][x].R := ((LUT[y + t].R - LUT[y].R) * F) shr 16;
          SL[y][x].g := ((LUT[y + t].G - LUT[y].G) * F) shr 16;
          SL[y][x].B := ((LUT[y + t].B - LUT[y].B) * F) shr 16;
        end;
      end;
    end;
  end;
  SetLength(LUT, 0);
end;

procedure fxBCS(Dest: TBitmap; B, C, S: integer);
var
  x, y, i, k, v, pv: Integer;
  ci1, ci2, ci3: integer;
  alpha: integer;
  a: double;
  BrightnesLut: array[0..255] of byte;
  ContrastLut: array[0..255] of byte;
  BCLut: array[0..255] of byte;
  SaturationLut: record
    Grays: array[0..767] of Integer;
    Alpha: array[Byte] of Word;
  end;
  SLDest, DestOff: integer;
  p: pRGB24;
  s1, s2: integer;
  bool: boolean;
  color: tRGB24;
begin
  if c = 100 then
    c := 99;
  for i := 0 to 255 do
  begin
    if c > 0 then
      a := 1 / cos(c * (PI / 200))
    else
      a := 1 * cos(c * (3.1416 / 200));
    v := Round(a * (i - 170) + 170);
    if v > 255 then
      v := 255
    else if v < 0 then
      v := 0;
    ContrastLut[i] := v;
  end;
  for i := 0 to 255 do
  begin
    alpha := b;
    k := 256 - alpha;
    v := (k + alpha * i) div 256;
    if v < 0 then
      v := 0
    else if v > 255 then
      v := 255;
    BCLut[i] := ContrastLut[v];
  end;
  x := 0;
  for i := 1 to 256 do
    SaturationLut.Alpha[i - 1] := (i * s) shr 8;
  for i := 0 to 255 do
  begin
    y := i - SaturationLut.Alpha[i];
    SaturationLut.Grays[x] := y;
    Inc(x);
    SaturationLut.Grays[x] := y;
    Inc(x);
    SaturationLut.Grays[x] := y;
    Inc(x);
  end;
  SLDest := Integer(dest.ScanLine[0]);
  DestOff := Integer(dest.ScanLine[1]) - SLDest;
  for y := 0 to Dest.Height - 1 do
  begin
    for x := 0 to Dest.Width - 1 do
    begin
      color := pRGBArray(SLDest)[x];
      v := color.R + color.G + color.B;
      ci1 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.B];
      ci2 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.G];
      ci3 := SaturationLut.Grays[v] + SaturationLut.Alpha[color.R];
      if ci1 < 0 then
        ci1 := 0
      else if ci1 > 255 then
        ci1 := 255;
      if ci2 < 0 then
        ci2 := 0
      else if ci2 > 255 then
        ci2 := 255;
      if ci3 < 0 then
        ci3 := 0
      else if ci3 > 255 then
        ci3 := 255;
      PRGBArray(SLDest)[x].B := BCLut[ci1];
      PRGBArray(SLDest)[x].G := BCLut[ci2];
      PRGBArray(SLDest)[x].R := BCLut[ci3];
    end;
    inc(SLDest, DestOff);
  end;
end;

procedure fxHistCalc(Src: TBitmap; var HistR, HistG, HistB: THistogram);
var
  RGB: PRGBArray;
  x, y: Integer;
begin
  for x := 0 to 255 do
  begin
    histR[x] := 0;
    histG[x] := 0;
    histB[x] := 0;
  end;
  for y := 0 to src.Height - 1 do
  begin
    RGB := src.ScanLine[y];
    for x := 0 to src.Width - 1 do
    begin
      inc(histR[RGB[x].R]);
      inc(histG[RGB[x].G]);
      inc(histB[RGB[x].B]);
    end;
  end;
end;

procedure fxHistEqu(Src: TBitmap; Z: Single);
type
  THistSingle = array[0..255] of Single;
var
  RGB: PRGBArray;
  x, y: integer;
  q1, q2, q3: Single;
  histR, HistG, HistB: Thistogram;
  Hist, VCumSumR, VCumSumG, VCumSumB: THistSingle;

  function CumSum(hist: THistSingle): THistSingle;
  var
    x: Byte;
    Temp: THistsingle;
  begin
    Temp[0] := Hist[0];
    for x := 1 to 255 do
      Temp[x] := Temp[x - 1] + Hist[x];
    CumSum := Temp;
  end;

begin
  fxHistCalc(src, histR, histG, histB);
  q1 := 0; // RED Channel
  for x := 0 to 255 do
  begin
    Hist[x] := power(HistR[x], z);
    q1 := q1 + Hist[x];
  end;
  vcumsumR := cumsum(Hist);
  q2 := 0; // GREEN Channel
  for x := 0 to 255 do
  begin
    Hist[x] := power(HistG[x], z);
    q2 := q2 + Hist[x];
  end;
  vcumsumG := cumsum(Hist);
  q3 := 0; // BLUE Channel
  for x := 0 to 255 do
  begin
    Hist[x] := power(HistB[x], z);
    q3 := q3 + Hist[x];
  end;
  vcumsumB := cumsum(Hist);
  for y := 0 to src.height - 1 do
  begin
    RGB := src.scanline[y];
    for x := 0 to src.width - 1 do
    begin
      RGB[x].R := Trunc((255 / q1) * vcumsumR[RGB[x].R]);
      RGB[x].G := Trunc((255 / q2) * vcumsumG[RGB[x].G]);
      RGB[x].B := Trunc((255 / q3) * vcumsumB[RGB[x].B]);
    end;
  end;
end;

procedure fxExposure(Src: TBitmap; K: Single);
var
  RGB: PRGBArray;
  i, x, y, RGBOffset: Integer;
  lut: array[0..255] of integer;
begin
  for i := 0 to 255 do
  begin
    if k < 0 then
      lut[i] := i - ((-Round((1 - Exp((i / -128) * (k / 128))) * 256) * (i xor
        255)) shr 8)
    else
      lut[i] := i + ((Round((1 - Exp((i / -128) * (k / 128))) * 256) * (i xor
        255)) shr 8);
    if lut[i] < 0 then
      lut[i] := 0
    else if lut[i] > 255 then
      lut[i] := 255;
  end;
  RGB := src.ScanLine[0];
  RGBOffset := Integer(src.ScanLine[1]) - Integer(RGB);
  for y := 0 to src.Height - 1 do
  begin
    for x := 0 to src.Width - 1 do
    begin
      RGB[x].R := LUT[RGB[x].R];
      RGB[x].G := LUT[RGB[x].G];
      RGB[x].B := LUT[RGB[x].B];
    end;
    RGB := PRGBArray(Integer(RGB) + RGBOffset);
  end;
end;

procedure fxGrayScale(Src: TBitmap; Kind: Byte);
var
  SCL: PRGBArray;
  AddL, g, x, y: Integer;
  c: TRGB24;
begin
  SCL:= Src.Scanline[0];
  AddL:= Integer(Src.Scanline[1]) - Integer(SCL);
  for y := 0 to Src.Height - 1 do
  begin
    for x := 0 to Src.Width - 1 do
    begin
      c:= SCL[x];
      case Kind of
        0 : if (c.B > c.R) and (c.B > c.G) then
              g:= (Max(Max(c.R, c.G), c.B) + Min(Min(c.R, c.G), c.B)) shr 1
            else
              g:= ((c.R * 21) + (c.G * 71) + (c.B * 8)) div 100;
        1 : g:= (c.R + c.G + c.B) div 3;
        2 : g:= (Max(Max(c.R, c.G), c.B) + Min(Min(c.R, c.G), c.B)) shr 1;
      else
        g:= 0;
      end;
      SCL[x].R:= g;
      SCL[x].G:= g;
      SCL[x].B:= g;
    end;
    SCL := PRGBArray(Integer(SCL) + AddL);
  end;
end;

procedure fxSpray(Src: TBitmap; Amount: Integer);
var
  i, j, x, y, w, h, Val: Integer;
  PScanLine: array of PRGBArray;
begin
  h := Src.height;
  w := Src.Width;
  SetLength(PScanLine, Src.Height);
  PScanLine[0]:= (Src.Scanline[0]);
  i := Integer(Src.Scanline[1]) - Integer(PScanLine[0]);
  for y := 1 to Src.Height - 1 do
    PScanLine[y]:= PRGBArray(Integer(PScanLine[y - 1]) + i);
  for i := 0 to w - 1 do
    for j := 0 to h - 1 do
    begin
      Val := Random(Amount);
      x := i + Val - Random(Val * 2);
      y := j + Val - Random(Val * 2);
      if (x > -1) and (x < w) and (y > -1) and (y < h) then
        PScanLine[j][i]:= PScanLine[y][x];
    end;
end;

procedure fxSplitLight(Src: TBitmap; Amount: Integer);
var
  x, y, i: integer;
  SCL: PRGBArray;
  Lut: array[0..255] of byte;
begin
  for i := 0 to 255 do
    Lut[i] := variant(Sin(i / 255 * pi / 2) * 255);
  for i := 1 to amount do
    for y := 0 to src.height - 1 do
    begin
      SCL := src.scanline[y];
      for x := 0 to src.width - 1 do
      begin
        SCL[x].R := Lut[SCL[x].R];
        SCL[x].G := Lut[SCL[x].G];
        SCL[x].B := Lut[SCL[x].B];
      end;
    end;
end;

procedure fxSplitBlur(Src: TBitmap; Amount: Integer);
var
  x, y, x1, x2, y1, y2: integer;
  p: PRGB24;
  SCL: PRGBArray;
  ptrS: integer;
begin
  if Amount = 0 then
    Exit;
  SCL := src.ScanLine[0];
  ptrS := integer(src.ScanLine[1]) - integer(SCL);
  for y := 0 to src.Height - 1 do
  begin
    if y - amount < 0 then
      y1 := y
    else
      y1 := y - amount;
    if y + amount < src.Height then
      y2 := y + amount
    else
      y2 := src.Height - y;
    for x := 0 to src.Width - 1 do
    begin
      if x - amount < 0 then
        x1 := x
      else
        x1 := x - amount;
      if x + amount < src.Width then
        x2 := x + amount
      else
        x2 := src.Width - x;
      p := pRGB24(PtrS * y + (x * 3) + Integer(SCL));
      p.r := (pRGB24(PtrS * y1 + (x1 * 3) + Integer(SCL)).R +
        pRGB24(PtrS * y2 + (x1 * 3) + Integer(SCL)).R +
        pRGB24(PtrS * y1 + (x2 * 3) + Integer(SCL)).R +
        pRGB24(PtrS * y2 + (x2 * 3) + Integer(SCL)).R) shr 2;
      p.g := (pRGB24(PtrS * y1 + (x1 * 3) + Integer(SCL)).G +
        pRGB24(PtrS * y2 + (x1 * 3) + Integer(SCL)).G +
        pRGB24(PtrS * y1 + (x2 * 3) + Integer(SCL)).G +
        pRGB24(PtrS * y2 + (x2 * 3) + Integer(SCL)).G) shr 2;
      p.b := (pRGB24(PtrS * y1 + (x1 * 3) + Integer(SCL)).B +
        pRGB24(PtrS * y2 + (x1 * 3) + Integer(SCL)).B +
        pRGB24(PtrS * y1 + (x2 * 3) + Integer(SCL)).B +
        pRGB24(PtrS * y2 + (x2 * 3) + Integer(SCL)).B) shr 2;
    end;
  end;
end;

procedure fxSplitColor(Src: TBitmap; Amount: Integer);
type
  pRGBTripleArray = ^TRGBTripleArray;
  TRGBTripleArray = ARRAY[0..0] OF TRGBTriple;
var
  x,ry,by,gy,ry1,by1,gy1,y1 : integer;
  y,r,g,b,r1,b1,g1,i : integer;
  row, rowd : pRGBTripleArray;
  lut : array[0..255] of byte;
begin
  for i:= 0 to 255 do lut[i]:= variant(Sin(i/255*pi/2)*255);
  for y1 := 0 to src.Height-1 do
  begin
    row := src.Scanline[y1];
    rowd := src.Scanline[y1];
    for x := 0 to src.Width-1 do
    begin
      r1 := row[x].rgbtRed;
      g1 := row[x].rgbtGreen;
      b1 := row[x].rgbtBlue;
      ry1:= Round((70*r1-59*g1-11*b1)/100);
      by1:= Round((-30*r1-59*g1+89*b1)/100);
      gy1:= Round((-30*r1+41*g1-11*b1)/100);
      y  := Lut[Round((30*r1+59*g1+11*b1)/100)];
      by := Round((by1*amount)/100);
      ry := Round((ry1*amount)/100);
      gy := Round((gy1*amount)/100);
      r  := ry+y; g :=gy+y; b:=by+y;
      if r < 0 then r := 0 else if r > 255 then r := 255;
      if g < 0 then g := 0 else if g > 255 then g := 255;
      if b < 0 then b := 0 else if b > 255 then b := 255;
      with rowd[x] do
      begin
        rgbtRed := r;
        rgbtGreen := g;
        rgbtBlue := b;
      end;
    end;
  end;
end;

procedure fxMosaic(Src: TBitmap; Amount: Integer);
var
  x, y, i, j: integer;
  SCL, SCL1: PRGBArray;
  r, g, b: byte;
begin
  y := 0;
  repeat
    SCL := src.scanline[y];
    repeat
      j := 1;
      repeat
        SCL1 := src.scanline[y];
        x := 0;
        repeat
          r := SCL[x].R;
          g := SCL[x].G;
          b := SCL[x].B;
          i := 1;
          repeat
            SCL1[x].R := r;
            SCL1[x].G := g;
            SCL1[x].B := b;
            inc(x);
            inc(i);
          until (x >= src.width) or (i > Amount);
        until x >= src.width;
        inc(j);
        inc(y);
      until (y >= src.height) or (j > Amount);
    until (y >= src.height) or (x >= src.width);
  until y >= src.height;
end;

procedure fxHorzRoll(Src: TBitmap; Amount: Integer);
var
  Line: PRGBArray;
  p, y, n: Integer;
begin
  if Amount > Src.Width then
    Amount := Amount mod Src.Width;
  if Amount = 0 then
    Exit;
  n:= (Src.Width - Amount) * 3;
  Amount:= Amount * 3;
  GetMem(Line, Amount);
  for y := 0 to Src.Height - 1 do
  begin
    p := Integer(Src.Scanline[y]);
    CopyMemory(Line, Pointer(p + n), Amount);
    MoveMemory(Pointer(p + Amount), Pointer(p), n);
    CopyMemory(Pointer(p), Line, Amount);
  end;
  FreeMem(Line, Amount);
end;

procedure fxRoll(Src: TBitmap; RollH, RollV: Integer);
var
  Bmp: TBitmap;
  S, D: TRect;
begin
  Bmp:= TBitmap.Create;
  Bmp.PixelFormat:= Src.PixelFormat;
  Bmp.Width:= Src.Width;
  Bmp.Height:= Src.Height;
  if RollV > 0 then
  begin
    if RollV > Src.Height then
      RollV := RollV mod Src.Height;
    S:= Rect(0, 0, Bmp.Width, Bmp.Height - RollV);
    D:= Rect(0, RollV, Bmp.Width, Bmp.Height);
    Bmp.Canvas.CopyRect(D, Src.Canvas, S);
    S:= Rect(0, Bmp.Height - RollV, Bmp.Width, Bmp.Height);
    D:= Rect(0, 0, Bmp.Width, RollV);
    Bmp.Canvas.CopyRect(D, Src.Canvas, S);
    Src.Assign(Bmp);
  end;
  if RollH > 0 then
  begin
    if RollH > Src.Width then
      RollH := RollH mod Src.Width;
    S:= Rect(0, 0, Bmp.Width - RollH, Bmp.Height);
    D:= Rect(RollH, 0, Bmp.Width, Bmp.Height);
    Bmp.Canvas.CopyRect(D, Src.Canvas, S);
    S:= Rect(Bmp.Width - RollH, 0, Bmp.Width, Bmp.Height);
    D:= Rect(0, 0, RollH, Bmp.Height);
    Bmp.Canvas.CopyRect(D, Src.Canvas, S);
    Src.Assign(Bmp);
  end;
  Bmp.Free;
end;

procedure fxAutoYHist(Src: TBitmap; Z: Single);
type
  THistSingle = array[0..255] of Single;
var
  p0: PRGBArray;
  x, y: integer;
  q1: Single;
  histY: THistogram;
  Hist, VCumSumR: THistSingle;
  cy, ccr, ccb: byte;

  procedure HistCalc(src: tBitmap; var histY: THistogram);
  var
    p0: PRGBArray;
    sx, sy, x, y: Integer;
  begin
    sx := src.width - 1;
    sy := src.height - 1;
    for y := 0 to sy do
    begin
      p0 := src.ScanLine[y];
      for x := 0 to sx do
        inc(histY[RGBtoYi(p0[x])]);
    end;
  end;

  function CumSum(hist: THistSingle): THistSingle;
  var
    x: Byte;
    Temp: THistsingle;
  begin
    Temp[0] := Hist[0];
    for x := 1 to 255 do
      Temp[x] := Temp[x - 1] + Hist[x];
    CumSum := Temp;
  end;

begin
  HistCalc(src, histY);

  // Intensity Channel Y
  q1 := 0;
  for x := 0 to 255 do
  begin
    Hist[x] := power(HistY[x], z);
    q1 := q1 + Hist[x];
  end;
  vcumsumR := cumsum(Hist);

  for y := 0 to src.height - 1 do
  begin
    p0 := src.scanline[y];
    for x := 0 to src.width - 1 do
      with p0[x] do
      begin
        RGBToYCbCr(R, G, B, cy, ccb, ccr);
        cy := Trunc((255 / q1) * vcumsumR[cy]);
        YCbCrtoRGB(cy, ccb, ccr, R, G, B);
      end;
  end;
end;

procedure fxAutoContrast(Src : TBitmap);
var
  x,y : Integer;
  cY,cCb,cCr : byte;
  i,ic,ir : integer;
  ltresh, htresh : integer;
  tc : integer;
  p0 : PRGBArray;
  lut : array[0..255] of byte;
  hist : array[0..255] of integer;
  scaleF: Single;
begin
  ic:=src.Width-1;
  ir:=src.Height-1;

  for i := 0 to 255 do hist[i] := 0;

  for y := 0 to ir do
  begin
    p0 := src.ScanLine[y];
    for x := 0 to ic do begin
      with p0[x] do
	      RGBtoYCbCr(R, G, B,cY,cCb,cCr);
      inc(hist[cy]);
    end;
  end;

  ltresh := 0;
  i := 0;
  while i < 256  do
  begin
    if hist[i] <> 0 then begin ltresh := i; i := 256; end;
    inc(i);
  end;

  htresh := 255;
  i := 255;
  while i >= 0 do
  begin
    if hist[i] <> 0 then begin htresh := i; i := -1; end;
    dec(i);
  end;

  for i := 0 to ltresh do lut[i]:= 0;
  for i := 255 downto htresh do lut[i]:= 255;

  scaleF := 255/(htresh-ltresh);
  for i:= ltresh to htresh do begin
    tc := Round((i-ltresh) * scaleF);
    if tc < 0 then tc := 0 else if tc > 255 then tc := 255;
    lut[i] := tc;
  end;

  for y := 0 to ir do begin
    p0 := src.ScanLine[y];
    for x := 0 to ic do begin
      with p0[x] do begin
	      RGBtoYCbCr(R, G, B,cY,cCb,cCr);
	      YCbCrToRGB(Lut[cY],cCb,cCr,R,G,B);
      end;
    end;
  end;
end;

procedure fxQFilter(src, dst: tBitmap; v : integer);
const
  F00 : array[0..9] of integer = ( 0, 0, 0,  0, 1, 0,  0, 0, 0,  1); // neutre
  F01 : array[0..9] of integer = ( 0, 0, 0,  0, 1, 1,  0, 1, 1,  4); // anti Alias
  F02 : array[0..9] of integer = ( 1, 2, 1,  2,20, 2,  1, 2, 1, 32); // Adoucir
  F22 : array[0..9] of integer = ( 2, 1, 2,  1,20, 1,  2, 1, 2, 32); // Adoucir bis
  F04 : array[0..9] of integer = (-1,-2,-1, -2,28,-2, -1,-2,-1, 16); // Durcir
  F44 : array[0..9] of integer = (-2,-1,-2, -1,28,-1, -2,-1,-2, 16); // Durcir bis
  F05 : array[0..9] of integer = ( 0,-1, 0, -1, 6,-1,  0,-1, 0,  2); // Durcir+
  F06 : array[0..9] of integer = (-5,-5,-5, -5,42,-5, -5,-5,-5,  2); // Lithographie
  F07 : array[0..9] of integer = (-1,-1,-1, -1, 9,-1, -1,-1,-1,  1); // Hi pass
  F08 : array[0..9] of integer = ( 0,-1, 0, -1, 1, 1,  0, 1, 0,  1); // relief
  F09 : array[0..9] of integer = ( 0, 0, 0,  0, 1, 0,  0, 0, 0,  1); // estamper
  F10 : array[0..9] of integer = ( 1, 1, 1,  1,-8, 1,  1, 1, 1,  1); // traits
  F11 : array[0..9] of integer = ( 3, 3, 3,  3,-25,3,  3, 3, 3,  1); // contour
  F12 : array[0..9] of integer = ( 0,-1,-1,  0, 2, 0,  0, 0, 0,  1); // sculpture
  F13 : array[0..9] of integer = ( 1, 1, 1,  1, 1, 1,  1, 1, 1,  9); // Smooth
  F14 : array[0..9] of integer = ( 3, 3, 3,  3, 8, 3,  3, 3, 3, 32); // Smooth
  F15 : array[0..9] of integer = ( 1, 2, 1,  2, 4, 2,  1, 2, 1, 16); // Smooth
  F16 : array[0..9] of integer = ( 0, 1, 0,  1,-4, 1,  0, 1, 0, 1);
  F17 : array[0..9] of integer = ( 0,-1, 0, -1,10,-1,  0,-1, 0, 6);
  F18 : array[0..9] of integer = (-1, 0,-1,  0, 7, 0, -1, 0,-1, 3);
type
  ByteArray = array[0..0] of byte;
  PByteArray = ^ByteArray;
var
  i,j,x,y,r,g,b: integer;
  s0,s1,s2,d0 : PByteArray;
  f : Array[0..9] of integer;
begin
  case v of
    1  : for i:= 0 to 9 do f[i]:= F00[i];
    2  : for i:= 0 to 9 do f[i]:= F01[i];
    3  : for i:= 0 to 9 do f[i]:= F02[i];
    4  : for i:= 0 to 9 do f[i]:= F22[i];
    5  : for i:= 0 to 9 do f[i]:= F04[i];
    6  : for i:= 0 to 9 do f[i]:= F44[i];
    7  : for i:= 0 to 9 do f[i]:= F05[i];
    8  : for i:= 0 to 9 do f[i]:= F06[i];
    9  : for i:= 0 to 9 do f[i]:= F07[i];
    10 : for i:= 0 to 9 do f[i]:= F08[i];
    11 : for i:= 0 to 9 do f[i]:= F09[i];
    12 : for i:= 0 to 9 do f[i]:= F10[i];
    13 : for i:= 0 to 9 do f[i]:= F11[i];
    14 : for i:= 0 to 9 do f[i]:= F12[i];
    15 : for i:= 0 to 9 do f[i]:= F13[i];
    16 : for i:= 0 to 9 do f[i]:= F14[i];
    17 : for i:= 0 to 9 do f[i]:= F15[i];
    18 : for i:= 0 to 9 do f[i]:= F16[i];
    19 : for i:= 0 to 9 do f[i]:= F17[i];
    20 : for i:= 0 to 9 do f[i]:= F18[i];
  end;

  for y := 1 to dst.Height-2 do begin
    s0 := src.ScanLine[y-1];
    s1 := src.ScanLine[y];
    s2 := src.ScanLine[y+1];
    d0 := dst.ScanLine[y];
    for x := 1 to dst.Width-2 do begin
      j:=x*3;
      r := (s0[j-3]*f[0])+(s0[j]*f[1])+(s0[j+3]*f[2])+
	   (s1[j-3]*f[3])+(s1[j]*f[4])+(s1[j+3]*f[5])+
	   (s2[j-3]*f[6])+(s2[j]*f[7])+(s2[j+3]*f[8]);
      r := r div f[9];
      if r < 0 then r := 0 else if r > 255 then r := 255;
      j:=j+1;
      g := (s0[j-3]*f[0])+(s0[j]*f[1])+(s0[j+3]*f[2])+
	   (s1[j-3]*f[3])+(s1[j]*f[4])+(s1[j+3]*f[5])+
	   (s2[j-3]*f[6])+(s2[j]*f[7])+(s2[j+3]*f[8]);
      g := g div f[9];
      if g < 0 then g := 0 else if g > 255 then g := 255;
      j:=j+1;
      b := (s0[j-3]*f[0])+(s0[j]*f[1])+(s0[j+3]*f[2])+
	   (s1[j-3]*f[3])+(s1[j]*f[4])+(s1[j+3]*f[5])+
	   (s2[j-3]*f[6])+(s2[j]*f[7])+(s2[j+3]*f[8]);
      b := b div f[9];
      if b < 0 then b := 0 else if b > 255 then b := 255;
      d0[j-2] := r;
      d0[j-1] := g;
      d0[j]   := b;
    end;
  end;
end;

procedure fxSoftFilter(Src: TBitmap; Amount: Integer);
var
  i,x,y,r1,g1,b1 : integer;
  tb : tbitmap;
  rows, rowd : PRGBArray;
begin
  tb := tbitmap.Create;
  tb.PixelFormat := pf24bit;
  tb.Width := src.Width;
  tb.Height := src.Height;
  tb.Assign(src);
  fxBoxBlur(tb,Amount, Amount, 3);
  for y := 0 to src.Height-1 do
  begin
    rows := tb.Scanline[y];
    rowd := Src.Scanline[y];
    for x:= 0 to src.Width-1 do
    begin
      rowd[x].r:= (rows[x].r+rowd[x].r) div 2;
      rowd[x].g:= (rows[x].g+rowd[x].g) div 2;
      rowd[x].b:= (rows[x].b+rowd[x].b) div 2;
    end;
  end;
  fxQFilter(Src,Src,3);
  fxQFilter(Src,Src,5);
  fxQFilter(Src,Src,4);
  fxQFilter(Src,Src,6);
  tb.Free;
end;


end.
